from __future__ import annotations

import enum
import logging
from dataclasses import dataclass
from typing import TYPE_CHECKING, List, Optional

from modelon.impact.client.configuration import Experimental
from modelon.impact.client.entities.workspace import (
    PublishedWorkspace,
    PublishedWorkspaceDefinition,
    PublishedWorkspaceType,
)
from modelon.impact.client.operations.orphan_cleanup import (
    OrphanPublishedWorkspaceOperation,
)

if TYPE_CHECKING:
    from modelon.impact.client.sal.service import Service

logger = logging.getLogger(__name__)


@dataclass
class OrphanPublishedWorkspaceOwner:
    id: str
    username: str
    group_name: str


@enum.unique
class PublishedWorkspaceAccessKind(enum.Enum):
    SHARED_BY_ME = "sharedByMe"
    SHARED_WITH_ME = "sharedWithMe"
    REQUESTED_BY_ME = "requestedByMe"
    REQUESTED_FROM_ME = "requestedFromMe"


@dataclass
class PublishedWorkspaceAccess:
    sharing_id: str
    requested_id: str
    requested_username: str
    published_workspace: Optional[PublishedWorkspace] = None


class PublishedWorkspacesClient:
    def __init__(self, service: Service):
        self._sal = service

    def find(
        self,
        *,
        name: str = "",
        first: int = 0,
        maximum: int = 20,
        has_data: bool = True,
        owner_username: str = "",
        type: Optional[PublishedWorkspaceType] = None,
        group_name: Optional[str] = None,
    ) -> List[PublishedWorkspace]:
        """Returns a list of published workspaces. The snapshots could be filtered based
        on the key-worded arguments.

        Args:
            name: Name of the workspace.
            first: Index of first matching resource to return.
            maximum: Maximum number of resources to return.
            has_data: If true, filters with
            status==PublishedWorkspaceUploadStatus.CREATED. If false
            returns everything.
            owner_username: If true, only workspaces published by the specified
            user are listed.
            type: Filter so only published workspace of a specified type are
                returned. If not given all published workspace types are
                returned.
            group_name: Group name to query published workspaces for. Only
                a user with impact-sys-admin role can query published workspaces
                shared with a group other than his own.

        Returns:
            A list of published workspace class objects.

        Example::

            pw_client = client.get_published_workspaces_client()
            pw_client.find()

        """
        data = self._sal.workspace.get_published_workspaces(
            name,
            first,
            maximum,
            has_data,
            owner_username,
            type.value if type else None,
            group_name,
        )["data"]["items"]
        return [
            PublishedWorkspace(
                item["id"],
                definition=PublishedWorkspaceDefinition.from_dict(item),
                service=self._sal,
            )
            for item in data
        ]

    def get_by_id(
        self, sharing_id: str, request_if_no_access: bool = False
    ) -> Optional[PublishedWorkspace]:
        """Returns the published workspace class object with the given ID.

        Args:
            sharing_id: ID of the published workspace.
            request_if_no_access: Request access if user doesn't have access.

        Returns:
            The published workspace class object.

        Example::

            pw_client = client.get_published_workspaces_client()
            pw_client.get("2h98hciwsniucwincj")

        """
        if request_if_no_access:
            self._sal.workspace.request_user_access(sharing_id)
            logger.info(
                "Access request sent for published workspaces with ID '%s'.", sharing_id
            )
            return None
        data = self._sal.workspace.get_published_workspace(sharing_id)
        definition = PublishedWorkspaceDefinition.from_dict(data)
        return PublishedWorkspace(data["id"], definition=definition, service=self._sal)

    @Experimental
    def get_by_access_kind(
        self,
        access_kind: PublishedWorkspaceAccessKind = PublishedWorkspaceAccessKind.SHARED_BY_ME,  # noqa
        first: int = 0,
        maximum: int = 20,
    ) -> List[PublishedWorkspaceAccess]:
        """Returns a list of PublishedWorkspaceAccess class objects. The snapshots could
        be filtered based on the key-worded arguments.

        Args:
            access_kind: Access kind for the published workspace.
            first: Index of first matching published workspace to return.
            maximum: Maximum number of published workspaces to return.

        Returns:
            A list of PublishedWorkspaceAccess class objects.

        Example::

            from modelon.impact.client.published_workspace_client import
                PublishedWorkspaceAccessKind

            pw_client = client.get_published_workspaces_client()
            pw_client.get_by_access_kind(PublishedWorkspaceAccessKind.REQUESTED_BY_ME)

        """
        data = self._sal.workspace.get_published_workspaces_by_kind(
            access_kind.value, first, maximum
        )["data"]["items"]
        return [
            PublishedWorkspaceAccess(
                item["sharingId"],
                item["requesterId"],
                item["requesterUsername"],
                PublishedWorkspace(
                    item["publishedWorkspace"]["id"],
                    definition=PublishedWorkspaceDefinition.from_dict(
                        item["publishedWorkspace"]
                    ),
                    service=self._sal,
                )
                if "publishedWorkspace" in item
                else None,
            )
            for item in data
        ]

    @Experimental
    def wipe_orphans(
        self, user: Optional[OrphanPublishedWorkspaceOwner] = None
    ) -> OrphanPublishedWorkspaceOperation:
        """Cleans up orphaned published workspaces.

        Args:
            user: User data for the creator of the orphaned published workspaces.
            Only a user with 'impact-workspace-auditor' role can specify a user data
            other than his own. Defaults to logged in user's data.

        Example::

            from modelon.impact.client import OrphanPublishedWorkspaceOwner

            pw_client = client.get_published_workspaces_client()
            orphans = pw_client.wipe_orphans().wait()

            user = OrphanPublishedWorkspaceOwner(
                id="8143248a-49ee-4473-a2ba-c721b9aafb42",
                username="user_1", group_name="impact-tenant-mycompany"
            )
            orphans = pwc.wipe_orphans(user=user).wait()

        """
        resp = self._sal.workspace.cleanup_orphans(user, only_list_orphans=False)
        return OrphanPublishedWorkspaceOperation(
            resp["data"]["location"], self._sal, "Orphan cleanup"
        )

    @Experimental
    def list_orphans(
        self, user: Optional[OrphanPublishedWorkspaceOwner] = None
    ) -> OrphanPublishedWorkspaceOperation:
        """Lists all orphaned published workspaces.

        Args:
            user: User data for the creator of the orphaned published workspaces.
            Only a user with 'impact-workspace-auditor' role can specify a user data
            other than his own. Defaults to logged in user's data.

        Example::

            from modelon.impact.client import OrphanPublishedWorkspaceOwner

            pw_client = client.get_published_workspaces_client()
            orphans = pw_client.list_orphans().wait()

            user = OrphanPublishedWorkspaceOwner(
                id="8143248a-49ee-4473-a2ba-c721b9aafb42",
                username="user_1", group_name="impact-tenant-mycompany"
            )
            orphans = pwc.list_orphans(user=user).wait()

        """
        resp = self._sal.workspace.cleanup_orphans(user, only_list_orphans=True)
        return OrphanPublishedWorkspaceOperation(
            resp["data"]["location"], self._sal, "Orphan listing"
        )



from __future__ import annotations

import enum
import logging
import time
from abc import abstractmethod
from typing import Any, Generic, Optional, Protocol, TypeVar, Union

from modelon.impact.client import exceptions

logger = logging.getLogger(__name__)
Entity = TypeVar("Entity")


class EntityFromOperation(Protocol[Entity]):
    def __call__(self, operation: BaseOperation[Entity], **kwargs: Any) -> Entity:
        ...


@enum.unique
class Status(enum.Enum):
    """Class representing an enumeration for the possible operation states."""

    PENDING = "pending"
    """Status for an operation that is pending."""

    RUNNING = "running"
    """Status for an operation that is running."""

    STOPPING = "stopping"
    """Status for an operation that has been cancelled and is stopping."""

    CANCELLED = "cancelled"
    """Status for an operation that has been cancelled."""

    DONE = "done"
    """Status for an operation that has been completed."""


@enum.unique
class AsyncOperationStatus(enum.Enum):
    """Defines all states for import."""

    RUNNING = "running"
    """Status for an async operation that is running."""

    READY = "ready"
    """Status for an async operation that is ready."""

    ERROR = "error"
    """Status for an async operation that has errors."""

    def done(self) -> bool:
        return self in [AsyncOperationStatus.READY, AsyncOperationStatus.ERROR]


class BaseOperation(Generic[Entity]):
    """Abstract base operation class."""

    def __init__(self, create_entity: EntityFromOperation):
        self._create_entity = create_entity

    @abstractmethod
    def data(self) -> Entity:
        """Returns the Entity class."""
        pass

    @property
    @abstractmethod
    def status(self) -> Any:
        """Returns the operation status as an enumeration."""
        pass

    @abstractmethod
    def cancel(self) -> Any:
        """Terminates the operation."""
        pass

    @property
    @abstractmethod
    def name(self) -> str:
        """Name of the operation."""
        pass

    @abstractmethod
    def wait(self, timeout: Optional[float]) -> Entity:
        """Waits for the operation to finish."""
        pass


class AsyncOperation(BaseOperation[Entity]):
    """File operation class containing base functionality."""

    def wait(self, timeout: Optional[float] = None) -> Entity:
        """Waits until the operation completes. Returns the operation class instance if
        operation completes.

        Args:
            timeout: Time to wait in seconds for achieving the status. By default
                the timeout is set to 'None', which signifies an infinite time
                to wait until the status is achieved.

        Returns:

            Entity class instance if operation completes.

        Raises:
            OperationTimeOutError if time exceeds set timeout.

        Example::

            workspace.upload_result('A.mat').wait(timeout = 120)

        """
        start_t = time.time()
        while True:
            logger.info(f"{self.name} in progress! Status : {self.status.name}")
            if self.status.done():
                logger.info(f"{self.name} completed! Status : {self.status.name}")
                return self.data()

            current_t = time.time()
            if timeout and current_t - start_t > timeout:
                raise exceptions.OperationTimeOutError(
                    current_status_name=self.status.name, timeout=timeout
                )

            time.sleep(0.5)

    def cancel(self) -> None:
        raise NotImplementedError("Cancel is not supported for this operation")


class ExecutionOperation(BaseOperation[Entity]):
    """Execution operation class containing base functionality."""

    def is_complete(self) -> bool:
        """Returns True if the operation has completed.

        Returns:
            True, if operation has completed.False, if operation has
            not completed.

        Example::

           model.compile(options).is_complete()
           workspace.execute(definition).is_complete()

        """
        return self.status == Status.DONE

    def wait(
        self,
        timeout: Optional[float] = None,
        status: Union[Status, tuple[Status, ...]] = (Status.DONE, Status.CANCELLED),
    ) -> Entity:
        """Waits until the operation achieves the set status. Returns the operation
        class instance if the set status is achieved.

        Args:
            timeout: Time to wait in seconds for achieving the status. By default
                the timeout is set to 'None', which signifies an infinite time
                to wait until the status is achieved.

            status:
                Operation status to be achieved.
                Default: Status.DONE

        Returns:

            Entity class instance if the set status is achieved.

        Raises:
            OperationTimeOutError if time exceeds set timeout.

        Example::

            model.compile(compile_options).wait(
                timeout = 120,
                status = Status.CANCELLED
                )
            workspace.execute(experiment_definition).wait(timeout = 120)

        """
        start_t = time.time()
        status_tuple = status if isinstance(status, tuple) else (status,)

        try:
            while True:
                logger.info(f"{self.name} in progress! Status : {self.status.name}")
                if self.status in status_tuple:
                    logger.info(f"{self.name} completed! Status : {self.status.name}")
                    return self.data()

                current_t = time.time()
                if timeout and current_t - start_t > timeout:
                    raise exceptions.OperationTimeOutError(
                        timeout=timeout,
                        current_status_name=self.status.name,
                    )

                time.sleep(0.5)
        except KeyboardInterrupt:
            logger.info("Execution cancelled!")
            self.cancel()
            raise



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict

from modelon.impact.client import exceptions
from modelon.impact.client.operations.base import (
    AsyncOperation,
    AsyncOperationStatus,
    Entity,
)

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import EntityFromOperation
    from modelon.impact.client.sal.service import Service

logger = logging.getLogger(__name__)


class FMUImportOperation(AsyncOperation[Entity]):
    """An operation class for the Model class."""

    def __init__(
        self,
        location: str,
        workspace_id: str,
        project_id: str,
        service: Service,
        create_entity: EntityFromOperation,
    ):
        super().__init__(create_entity)
        self._location = location
        self._workspace_id = workspace_id
        self._project_id = project_id
        self._sal = service
        self._create_entity = create_entity

    def __repr__(self) -> str:
        return f"FMU import operations for id '{self.id}'"

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, FMUImportOperation) and obj._location == self._location

    @property
    def id(self) -> str:
        """FMU import id."""
        return self._location.split("/")[-1]

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return "Model import"

    def _info(self) -> Dict[str, Any]:
        return self._sal.imports.get_import_status(self._location)["data"]

    def data(self) -> Entity:
        """Returns a new Model class instance.

        Returns:
            A Model class instance.

        """
        info = self._info()
        if info["status"] == AsyncOperationStatus.ERROR.value:
            raise exceptions.IllegalFMUImport(
                f"FMU import failed! Cause: {info['error'].get('message')}"
            )
        if info.get("importWarnings"):
            logger.warning(f"Import Warnings: {'. '.join(info['importWarnings'])}")

        class_name = info["data"]["fmuClassPath"]
        return self._create_entity(
            self,
            class_name=class_name,
            workspace_id=self._workspace_id,
            project_id=self._project_id,
        )

    @property
    def status(self) -> AsyncOperationStatus:
        """Returns the upload status as an enumeration.

        Returns:
            The AsyncOperationStatus enum. The status can have the enum values
            AsyncOperationStatus.READY, AsyncOperationStatus.RUNNING or
            AsyncOperationStatus.ERROR

        Example::

            model.import_fmu('test.fmu').status

        """
        return AsyncOperationStatus(self._info()["status"])



from __future__ import annotations

import os
from typing import TYPE_CHECKING, Any, Dict

from modelon.impact.client import exceptions
from modelon.impact.client.operations.base import (
    AsyncOperation,
    AsyncOperationStatus,
    BaseOperation,
    Entity,
)

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import EntityFromOperation
    from modelon.impact.client.sal.exports import ExportService
    from modelon.impact.client.sal.service import Service


class Export:
    def __init__(self, export_service: ExportService, download_uri: str):
        self._export_sal = export_service
        self._download_uri = download_uri

    @property
    def id(self) -> str:
        """ID of the export."""
        return self._download_uri.split("/")[-1]

    def download_as(self, path_to_download: str) -> str:
        """Writes the binary archive to a file. Returns the path to downloaded archive.

        Args:
            path_to_download: The path to store the downloaded workspace.

        Returns:
            local path to the downloaded archive.

        Example::

            path = workspace.export().wait().download_as('/home/workspace.zip')
            path = workspace.export().wait().download_as('workspace.zip')

        """
        data = self._export_sal.export_download(self._download_uri)
        os.makedirs(os.path.dirname(path_to_download), exist_ok=True)
        with open(path_to_download, "wb") as f:
            f.write(data)
        return path_to_download

    @classmethod
    def from_operation(cls, operation: BaseOperation[Export], **kwargs: Any) -> Export:
        assert isinstance(operation, WorkspaceExportOperation)
        return cls(**kwargs, export_service=operation._sal.exports)


class WorkspaceExportOperation(AsyncOperation[Entity]):
    """An export operation class for the Workspace class."""

    def __init__(
        self, location: str, service: Service, create_entity: EntityFromOperation
    ):
        super().__init__(create_entity)
        self._location = location
        self._sal = service

    def __repr__(self) -> str:
        return f"Workspace export operations for id '{self.id}'"

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, WorkspaceExportOperation)
            and obj._location == self._location
        )

    @property
    def id(self) -> str:
        """Workspace export id."""
        return self._location.split("/")[-1]

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return "Workspace export"

    def _info(self) -> Dict[str, Any]:
        return self._sal.exports.get_export_status(self._location)["data"]

    def data(self) -> Entity:
        """Returns a Export class instance.

        Returns:
            An Export class instance.

        """
        info = self._info()
        if info["status"] == AsyncOperationStatus.ERROR.value:
            raise exceptions.IllegalWorkspaceExport(
                f"Workspace export failed! Cause: {info['error'].get('message')}"
            )
        return self._create_entity(self, download_uri=info["data"]["downloadUri"])

    @property
    def status(self) -> AsyncOperationStatus:
        """Returns the upload status as an enumeration.

        Returns:
            The AsyncOperationStatus enum. The status can have the enum values
            AsyncOperationStatus.READY, AsyncOperationStatus.RUNNING or
            AsyncOperationStatus.ERROR

        Example::

            workspace.download().status

        """
        return AsyncOperationStatus(self._info()["status"])



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, List, Optional

from modelon.impact.client.entities.file_uri import FileURI, get_resource_URI_from_str
from modelon.impact.client.entities.interfaces.case import CaseInterface, CaseReference
from modelon.impact.client.entities.interfaces.custom_function import (
    CustomFunctionInterface,
)
from modelon.impact.client.entities.interfaces.experiment import (
    ExperimentInterface,
    ExperimentReference,
)
from modelon.impact.client.options import ProjectExecutionOptions
from modelon.impact.client.sal.service import Service

if TYPE_CHECKING:
    from modelon.impact.client.options import (
        CompilerOptions,
        RuntimeOptions,
        SimulationOptions,
        SolverOptions,
    )

logger = logging.getLogger(__name__)


class _Parameter:
    __slots__ = ["_name", "_value", "_value_type", "_valid_values"]

    _JSON_2_PY_TYPE = {
        "Number": (
            float,
            int,
        ),
        "String": (str,),
        "Boolean": (bool,),
        "Enumeration": (str,),
        "CaseResult": (CaseInterface,),
        "ExperimentResult": (ExperimentInterface,),
        "FileURI": (FileURI,),
        "VariableNames": (list,),
    }

    def __init__(self, name: str, value: Any, value_type: str, valid_values: List[Any]):
        self._name = name
        self._value = value
        self._value_type = value_type
        self._valid_values = valid_values

    @property
    def name(self) -> str:
        return self._name

    @property
    def type(self) -> str:
        return self._value_type

    @property
    def value(self) -> Any:
        return self._value

    @value.setter
    def value(self, value: Any) -> None:
        if not isinstance(value, self._JSON_2_PY_TYPE[self._value_type]):
            raise ValueError(
                f"Cannot set {self._name} to {value}, its type is {self._value_type}"
            )

        if self._value_type == "Enumeration" and value not in self._valid_values:
            raise ValueError(
                f"Cannot set enumeration '{self._name}' to '{value}', "
                f"must be one of {self._valid_values}"
            )
        self._value = value


class ParameterDict(dict):
    def as_raw_dict(self) -> Dict[str, Any]:
        new_dict = {}
        for key, value in self.items():
            if isinstance(value, ExperimentInterface):
                new_dict[key] = value.id
            elif isinstance(value, CaseInterface):
                new_dict[key] = f"{value.experiment_id}/{value.id}"
            elif isinstance(value, FileURI):
                new_dict[key] = str(value)
            else:
                new_dict[key] = value
        return new_dict


class CustomFunction(CustomFunctionInterface):
    """Class containing CustomFunction functionalities."""

    def __init__(
        self,
        workspace_id: str,
        name: str,
        parameter_data: List[Dict[str, Any]],
        service: Service,
    ):
        self._name = name
        self._workspace_id = workspace_id
        self._parameter_data = parameter_data
        self._param_by_name = {
            p["name"]: _Parameter(
                p["name"],
                p.get("defaultValue", ""),
                p["type"],
                p.get("values", []),
            )
            for p in parameter_data
        }
        self._sal = service

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, CustomFunction) and obj._name == self._name

    def __repr__(self) -> str:
        return f"Custom function '{self._name}'"

    @property
    def name(self) -> str:
        "Custom function name"
        return self._name

    def with_parameters(self, **modified: Any) -> CustomFunction:
        """Sets/updates the custom_function parameters for an experiment.

        Args:
            parameters: A key-worded, variable-length argument list of custom_function
                parameters.

        Example::

            custom_function.with_parameters(start_time=0.0, final_time=2.0)

        """
        new = CustomFunction(
            self._workspace_id, self._name, self._parameter_data, self._sal
        )
        for name, value in modified.items():
            if name not in new._param_by_name:
                raise ValueError(
                    f"The custom function '{self._name}' "
                    f"does not have a parameter '{name}'"
                )

            parameter = new._param_by_name[name]
            if parameter.type == "ExperimentResult" and isinstance(value, str):
                exp_info = self._sal.workspace.experiment_get(self._workspace_id, value)
                parameter.value = ExperimentReference(
                    self._workspace_id, value, self._sal, exp_info
                )
            elif parameter.type == "CaseResult" and isinstance(value, str):
                experiment_id, case_id = value.split("/")
                case_info = self._sal.experiment.case_get(
                    self._workspace_id, experiment_id, case_id
                )
                parameter.value = CaseReference(
                    case_id, self._workspace_id, experiment_id, self._sal, case_info
                )
            elif parameter.type == "FileURI" and isinstance(value, str):
                parameter.value = get_resource_URI_from_str(value)
            else:
                parameter.value = value

        return new

    @property
    def parameter_values(self) -> ParameterDict:
        """Custom_function parameters and value as a dictionary."""
        return ParameterDict({p.name: p.value for p in self._param_by_name.values()})

    def get_options(
        self, use_defaults: Optional[bool] = False
    ) -> ProjectExecutionOptions:
        """Get project execution option.

        Args:
            use_defaults:
                If true, default compiler options are used.

        Example::

            opts = custom_function.get_options()
            comp_opts = opts.compiler_options
            comp_opts_mod = copts.with_values(include_protected_variables=False)

        """
        if use_defaults:
            options = self._sal.custom_function.custom_function_default_options_get(
                self._workspace_id, self._name
            )
        else:
            options = self._sal.custom_function.custom_function_options_get(
                self._workspace_id, self._name
            )
        return ProjectExecutionOptions(options, self.name)

    def get_compiler_options(self, use_defaults: bool = False) -> CompilerOptions:
        """Return a modelon.impact.client.options.CompilerOptions object.

        Args:
            use_defaults:
                If True, default compiler options are used.

        Returns:
            A CompilerOptions object.

        Example::

            opts = custom_function.get_compiler_options()
            compiler_options = opts.with_values(include_protected_variables=False)

        """
        return self.get_options(use_defaults=use_defaults).compiler_options

    def get_runtime_options(self, use_defaults: bool = False) -> RuntimeOptions:
        """Return a RuntimeOptions object.

        Args:
            use_defaults:
                If True, default compiler options are used.

        Returns:
            A RuntimeOptions object.

        Example::

            opts = custom_function.get_runtime_options()
            opts_2 = opts.with_values(cs_solver=0)

        """
        return self.get_options(use_defaults=use_defaults).runtime_options

    def get_solver_options(self, use_defaults: bool = False) -> SolverOptions:
        """Return a modelon.impact.client.options.SolverOptions object.

        Args:
            use_defaults:
                If True, default compiler options are used.

        Returns:
            A SolverOptions object.

        Example::

            opts = custom_function.get_solver_options()
            opts_2 = opts.with_values(rtol=1e-7)

        """
        return self.get_options(use_defaults=use_defaults).solver_options

    def get_simulation_options(self, use_defaults: bool = False) -> SimulationOptions:
        """Return a SimulationOptions object.

        Args:
            use_defaults:
                If True, default compiler options are used.

        Returns:
            A SimulationOptions object.

        Example::

            opts = custom_function.get_simulation_options()
            opts_2 = opts.with_values(ncp=500)

        """
        return self.get_options(use_defaults=use_defaults).simulation_options



"""Response class."""
import re
from typing import Any, Dict, Optional, Text

from modelon.impact.client.sal import exceptions


class ResponseError:
    def __init__(self, message: str, code: int):
        self.message = message
        self.code = code


class Response:
    def __init__(self, resp_obj: Any):
        self._resp_obj = resp_obj

    def _is_json(self) -> bool:
        return "application/json" in self._resp_obj.headers.get("content-type")

    @property
    def status_code(self) -> int:
        return self._resp_obj.status_code

    @property
    def ok(self) -> bool:
        return self._resp_obj.ok

    @property
    def error(self) -> ResponseError:
        if self._resp_obj.ok:
            raise ValueError("This request was successfull!")

        if not self._is_json():
            raise exceptions.ErrorBodyIsNotJSONError(
                f"Error response was not JSON: {self._resp_obj.content}"
            )

        json = self._resp_obj.json()
        if "error" not in json:
            raise exceptions.ErrorJSONInvalidFormatError(
                f"Error response JSON format unknown: {self._resp_obj.content}"
            )

        error = json["error"]
        return ResponseError(error["message"], error["code"])


class JSONResponse(Response):
    def __init__(self, resp_obj: Any):
        super().__init__(resp_obj)

    @property
    def data(self) -> Dict[str, Any]:
        if not self._resp_obj.ok:
            raise exceptions.HTTPError(self.error.message, self._resp_obj.status_code)

        if not self._is_json():
            raise exceptions.InvalidContentTypeError(
                "Incorrect content type on response, expected JSON"
            )

        return self._resp_obj.json()

    @property
    def headers(self) -> Dict[str, Any]:
        return self._resp_obj.headers


class TextResponse(Response):
    def __init__(self, resp_obj: Any):
        super().__init__(resp_obj)

    def _is_txt(self) -> bool:
        return "text/plain" in self._resp_obj.headers.get("content-type")

    @property
    def data(self) -> Text:
        if not self._resp_obj.ok:
            raise exceptions.HTTPError(self.error.message, self._resp_obj.status_code)

        if not self._is_txt():
            raise exceptions.InvalidContentTypeError(
                "Incorrect content type on response, expected text"
            )

        return self._resp_obj.text


class XMLResponse(Response):
    def __init__(self, resp_obj: Any):
        super().__init__(resp_obj)

    def _is_xml(self) -> bool:
        return "application/xml" in self._resp_obj.headers.get("content-type")

    @property
    def data(self) -> Text:
        if not self._resp_obj.ok:
            raise exceptions.HTTPError(self.error.message, self._resp_obj.status_code)

        if not self._is_xml():
            raise exceptions.InvalidContentTypeError(
                "Incorrect content type on response, expected XML"
            )

        return self._resp_obj.text


class ZIPResponse(Response):
    def __init__(self, resp_obj: Any):
        super().__init__(resp_obj)

    def _is_zip(self) -> bool:
        return "application/zip" in self._resp_obj.headers.get("content-type")

    @property
    def data(self) -> bytes:
        if not self._resp_obj.ok:
            raise exceptions.HTTPError(self.error.message, self._resp_obj.status_code)

        if not self._is_zip():
            raise exceptions.InvalidContentTypeError(
                "Incorrect content type on response, expected a Binary "
                "compressed archive"
            )

        return self._resp_obj.content


class FileResponse(Response):
    def __init__(self, resp_obj: Any, content_type: Optional[str] = None):
        super().__init__(resp_obj)
        self.content_type = content_type

    def _assert_expected_content_type(self) -> None:
        if self.content_type is None:
            return
        resp_content_type = self._resp_obj.headers.get("content-type")
        if self.content_type not in resp_content_type:
            raise exceptions.InvalidContentTypeError(
                f"Incorrect content type on response, expected {self.content_type}"
            )

    @property
    def stream(self) -> bytes:
        if not self._resp_obj.ok:
            raise exceptions.HTTPError(self.error.message, self._resp_obj.status_code)

        if self.content_type is None:
            self.content_type = self._resp_obj.headers.get("content-type")[0]
        else:
            self._assert_expected_content_type()

        return (
            self._resp_obj.text
            if self.content_type.startswith("text")
            else self._resp_obj.content
        )

    @property
    def headers(self) -> Dict[str, Any]:
        return self._resp_obj.headers

    @property
    def file_name(self) -> str:
        d = self.headers["content-disposition"]
        return re.findall("filename=(.+)", d)[0].strip('"')


class CSVResponse(FileResponse):
    def __init__(self, resp_obj: Any):
        super().__init__(resp_obj, "text/csv")


class OctetStreamResponse(FileResponse):
    def __init__(self, resp_obj: Any):
        super().__init__(resp_obj, "application/octet-stream")


class MatStreamResponse(FileResponse):
    def __init__(self, resp_obj: Any):
        super().__init__(resp_obj, "application/vnd.impact.mat.v1+octet-stream")



"""Service class."""
import logging
from typing import Any, Dict, Optional

from modelon.impact.client.sal import exceptions
from modelon.impact.client.sal.context import Context
from modelon.impact.client.sal.custom_function import CustomFunctionService
from modelon.impact.client.sal.experiment import ExperimentService
from modelon.impact.client.sal.exports import ExportService
from modelon.impact.client.sal.external_result import ExternalResultService
from modelon.impact.client.sal.http import HTTPClient
from modelon.impact.client.sal.imports import ImportService
from modelon.impact.client.sal.model_executable import ModelExecutableService
from modelon.impact.client.sal.project import ProjectService
from modelon.impact.client.sal.uri import URI
from modelon.impact.client.sal.users import UsersService
from modelon.impact.client.sal.workspace import WorkspaceService

logger = logging.getLogger(__name__)


def _get_impact_base_uri(uri: URI, http_client: HTTPClient) -> URI:
    url = (uri / "hub/api/user").resolve()
    server_url = http_client.get_json(url)["server"]
    return uri / f"{server_url}/impact"


def is_jupyterhub_url(uri: URI, context: Optional[Context] = None) -> bool:
    _JUPYTERHUB_VERSION_HEADER = "x-jupyterhub-version"
    url = (uri / "hub/api/").resolve()

    try:
        http_client = HTTPClient(context=context)
        response = http_client.get_json_response(url)
    except (
        exceptions.CommunicationError,
        exceptions.SSLError,
        exceptions.HTTPError,
        exceptions.ErrorBodyIsNotJSONError,
    ):
        return False
    except Exception:
        logger.warning(
            "Unknown exception trying to determine if URL is to JupyterHub or "
            "Modelon Impact, will assume the URL goes directly to the "
            "Modelon Impact API"
        )
        return False
    return _JUPYTERHUB_VERSION_HEADER in response.headers


class Service:
    def __init__(self, uri: URI, api_key: str, context: Optional[Context] = None):
        self._http_client = HTTPClient(api_key, context)
        self._base_uri = (
            _get_impact_base_uri(uri, self._http_client)
            if is_jupyterhub_url(uri, context)
            else uri
        )
        self.workspace = WorkspaceService(self._base_uri, self._http_client)
        self.project = ProjectService(self._base_uri, self._http_client)
        self.model_executable = ModelExecutableService(
            self._base_uri, self._http_client
        )
        self.experiment = ExperimentService(self._base_uri, self._http_client)
        self.custom_function = CustomFunctionService(self._base_uri, self._http_client)
        self.external_result = ExternalResultService(self._base_uri, self._http_client)
        self.users = UsersService(self._base_uri, self._http_client)
        self.exports = ExportService(self._base_uri, self._http_client)
        self.imports = ImportService(self._base_uri, self._http_client)

    def api_get_metadata(self) -> Dict[str, Any]:
        url = (self._base_uri / "api/").resolve()
        response = self._http_client.get_json_response(url)

        return response.data

    def get_executions(self) -> Dict[str, Any]:
        url = (self._base_uri / "api/executions").resolve()
        resp = self._http_client.get_json_response(url)
        return resp.data



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Dict, Optional

if TYPE_CHECKING:
    from modelon.impact.client.sal.service import Service


class ExperimentInterface(ABC):
    def __init__(
        self,
        workspace_id: str,
        exp_id: str,
        service: Service,
        info: Optional[Dict[str, Any]] = None,
    ):
        self._exp_id = exp_id
        self._workspace_id = workspace_id
        self._sal = service
        self._info = info

    @property
    @abstractmethod
    def id(self) -> str:
        """Experiment id."""
        raise NotImplementedError

    @abstractmethod
    def get_cases(self) -> Any:
        """Returns a list of case objects for an experiment."""
        raise NotImplementedError


class ExperimentReference(ExperimentInterface):
    def __init__(
        self,
        workspace_id: str,
        exp_id: str,
        service: Service,
        info: Optional[Dict[str, Any]] = None,
    ) -> None:
        super().__init__(workspace_id, exp_id, service, info)

    @property
    def id(self) -> str:
        """Experiment id."""
        return self._exp_id

    def get_cases(self) -> Any:
        """Returns a list of case objects for an experiment."""
        raise NotImplementedError



from __future__ import annotations

from abc import ABC
from collections.abc import Mapping
from copy import deepcopy
from typing import Any, Dict, Iterator, List


def _set_options(options: Dict[str, Any], **modified: Any) -> Dict[str, Any]:
    opts = deepcopy(options)
    for name, value in modified.items():
        opts[name] = value
    return opts


class BaseExecutionOptions(Mapping, ABC):
    """Base class for the simulation, compiler, solver and runtime options settings."""

    def __init__(self, values: Dict[str, Any], custom_function_name: str):
        self._values = values
        self._custom_function_name = custom_function_name

    def __repr__(self) -> str:
        return f"{type(self).__name__} for '{self._custom_function_name}'"

    def __getitem__(self, key: str) -> Any:
        return self._values[key]

    def __iter__(self) -> Iterator[Any]:
        return self._values.__iter__()

    def __len__(self) -> int:
        return self._values.__len__()

    def with_values(self, **modified: Any) -> Any:
        """Sets/updates the options.

        Args:
            parameters: A key-worded, variable-length argument list of options.

        Example::

            cmp_opts = custom_function.get_compiler_options().with_values(
                include_protected_variables=False)
            runtime_opts = custom_function.get_runtime_options().with_values(
                cs_solver=0)
            sol_opts = custom_function.get_solver_options().with_values(rtol=1e-7)
            sim_opts = custom_function.get_simulation_options().with_values(ncp=500)

        """
        values = _set_options(self._values, **modified)
        return self.__class__(values, self._custom_function_name)


class CompilerOptions(BaseExecutionOptions):
    pass


class RuntimeOptions(BaseExecutionOptions):
    pass


class SolverOptions(BaseExecutionOptions):
    pass


class SimulationOptions(BaseExecutionOptions):
    def with_result_filter(self, filter: List[str]) -> SimulationOptions:
        """Sets the variable filter for results.

        Args:
            filter:
                A list of filter patterns for choosing which variables to actually
                store result for. The syntax can be found in
                https://en.wikipedia.org/wiki/Glob_(programming).
                An example for filter is `*der`, which would stores all
                variables ending with `der`.

        Example::

            sim_opts = custom_function.get_simulation_options().with_result_filter(
                filter = ["*.phi"])

        """
        if not isinstance(filter, list) or [
            p for p in filter if not isinstance(p, str)
        ]:
            raise ValueError(
                "The filter must be specified as a list of pattern strings!"
            )
        return self.with_values(filter=str(filter))


class ProjectExecutionOptions:
    def __init__(self, data: Dict[str, Any], name: str):
        self._data = data
        self._name = name

    @property
    def custom_function(self) -> str:
        return self._name

    @property
    def compiler_options(self) -> CompilerOptions:
        return CompilerOptions(self._data.get("compiler", {}), self.custom_function)

    @property
    def runtime_options(self) -> RuntimeOptions:
        return RuntimeOptions(self._data.get("runtime", {}), self.custom_function)

    @property
    def simulation_options(self) -> SimulationOptions:
        return SimulationOptions(self._data.get("simulation", {}), self.custom_function)

    @property
    def solver_options(self) -> SolverOptions:
        return SolverOptions(self._data.get("solver", {}), self.custom_function)

    def to_dict(self) -> Dict[str, Any]:
        return {"customFunction": self._name, **self._data}



from __future__ import annotations

from typing import Any, Optional


class Error(Exception):
    pass


class UnsupportedSemanticVersionError(Error):
    pass


class NoAssignedLicenseError(Error):
    pass


class OperationTimeOutError(Error):
    def __init__(
        self,
        current_status_name: str,
        expected_status_names: Optional[list[str]] = None,
        timeout: Optional[float] = None,
    ) -> None:
        timeout_msg = f"Time exceeded the set timeout - {timeout}s. " if timeout else ""
        status_msg = f"Present status of operation is {current_status_name}. "
        expected_status_names_msg = (
            f"Expected statuses: {', '.join(expected_status_names)}."
            if expected_status_names
            else ""
        )
        super().__init__(timeout_msg + status_msg + expected_status_names_msg)


class OperationFailureError(Error):
    @classmethod
    def for_operation(cls, operation_name: str) -> OperationFailureError:
        return cls(
            f"{operation_name} was cancelled before completion! "
            f"Log file generated for cancelled {operation_name} is empty!"
        )


class OperationNotCompleteError(Error):
    @classmethod
    def for_operation(
        cls, operation_name: str, status: Any
    ) -> OperationNotCompleteError:
        return cls(
            f"{operation_name} is still in progress! Status : {status}."
            f" Please call the wait() method on the {operation_name} operation"
            " to wait until completion!"
        )


class OperationCompleteError(Error):
    pass


class FailedOrphanCleanup(Error):
    pass


class IllegalProjectImport(Error):
    pass


class IllegalWorkspaceImport(Error):
    pass


class IllegalWorkspaceExport(Error):
    pass


class IllegalWorkspaceConversion(Error):
    pass


class IllegalContentImport(Error):
    pass


class IllegalCustomArtifactImport(Error):
    pass


class IllegalCaseResultImport(Error):
    pass


class IllegalFMUImport(Error):
    pass


class ExternalResultUploadError(Error):
    pass


class NoSuchCustomArtifactError(Error):
    pass


class NoAssociatedPublishedWorkspaceError(Error):
    pass


class RemotePublishedWorkspaceLinkError(Error):
    pass


class AuthenticationError(Error):
    pass



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from modelon.impact.client import exceptions
from modelon.impact.client.operations.base import (
    AsyncOperation,
    AsyncOperationStatus,
    Entity,
)

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import EntityFromOperation
    from modelon.impact.client.sal.service import Service


class ExternalResultImportOperation(AsyncOperation[Entity]):
    """An operation class for the ExternalResult class."""

    def __init__(
        self, location: str, service: Service, create_entity: EntityFromOperation
    ):
        super().__init__(create_entity)
        self._location = location
        self._sal = service
        self._create_entity = create_entity

    def __repr__(self) -> str:
        return f"Result import operations for id '{self.id}'"

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, ExternalResultImportOperation)
            and obj._location == self._location
        )

    @property
    def id(self) -> str:
        """Result import id."""
        return self._location.split("/")[-1]

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return "Result import"

    def _info(self) -> Dict[str, Any]:
        return self._sal.imports.get_import_status(self._location)["data"]

    def data(self) -> Entity:
        """Returns a new ExternalResult class instance.

        Returns:
            A ExternalResult class instance.

        """
        info = self._info()
        if info["status"] == AsyncOperationStatus.ERROR.value:
            raise exceptions.ExternalResultUploadError(
                f"External result upload failed! Cause: {info['error'].get('message')}"
            )
        resp = self._sal.external_result.get_uploaded_result(self.id)
        return self._create_entity(self, result_id=resp["data"]["id"])

    @property
    def status(self) -> AsyncOperationStatus:
        """Returns the upload status as an enumeration.

        Returns:
            The AsyncOperationStatus enum. The status can have the enum values
            AsyncOperationStatus.READY, AsyncOperationStatus.RUNNING or
            AsyncOperationStatus.ERROR

        Example::

            workspace.upload_result('A.mat').status

        """
        return AsyncOperationStatus(self._info()["status"])



"""Model executable service module."""
from typing import Any, Dict, List, Tuple

from modelon.impact.client.sal.http import HTTPClient
from modelon.impact.client.sal.uri import URI


class ModelExecutableService:
    def __init__(self, uri: URI, http_client: HTTPClient):
        self._base_uri = uri
        self._http_client = http_client

    def fmu_setup(
        self, workspace_id: str, options: Dict[str, Any], get_cached: bool
    ) -> Tuple[str, List[Dict[str, Any]]]:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/model-executables"
            f"?getCached={'true' if get_cached else 'false'}"
        ).resolve()
        resp = self._http_client.post_json(url, body=options)
        return resp["id"], resp["parametersResolved"]

    def compile_model(self, workspace_id: str, fmu_id: str) -> str:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}/compilation"
        ).resolve()
        self._http_client.post_json_no_response_body(url)
        return fmu_id

    def compile_log(self, workspace_id: str, fmu_id: str) -> str:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}/compilation/"
            "log"
        ).resolve()
        return self._http_client.get_text(url)

    def model_description_get(self, workspace_id: str, fmu_id: str) -> str:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}/"
            "model-description"
        ).resolve()
        return self._http_client.get_xml(url)

    def fmu_delete(self, workspace_id: str, fmu_id: str) -> None:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}"
        ).resolve()
        self._http_client.delete_json(url)

    def compile_status(self, workspace_id: str, fmu_id: str) -> Dict[str, Any]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}/compilation"
        ).resolve()
        return self._http_client.get_json(url)

    def compile_cancel(self, workspace_id: str, fmu_id: str) -> None:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}/compilation"
        ).resolve()
        return self._http_client.delete_json(url)

    def settable_parameters_get(self, workspace_id: str, fmu_id: str) -> List[str]:
        url = (
            self._base_uri
            / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}/"
            "settable-parameters"
        ).resolve()
        return self._http_client.get_json(url)

    def ss_fmu_metadata_get(
        self, workspace_id: str, fmu_id: str, parameter_state: Dict[str, Any]
    ) -> Dict[str, Any]:
        url = (
            self._base_uri / f"api/workspaces/{workspace_id}/model-executables/{fmu_id}"
            "/steady-state-metadata"
        ).resolve()
        return self._http_client.post_json(url, body=parameter_state)



import abc
import enum
from typing import Any, Union

Scalar = Union[bool, int, float, str]


class DataType(enum.Enum):
    BOOLEAN = "BOOLEAN"
    INTEGER = "INTEGER"
    REAL = "REAL"
    STRING = "STRING"
    ENUMERATION = "ENUMERATION"


class Modifier(abc.ABC):
    @abc.abstractmethod
    def to_dict(self, name: str) -> dict[str, Any]:
        """Returns a dict representation of the modifier."""

    @abc.abstractmethod
    def to_value(self) -> Scalar:
        """Returns a single scalar representing the modifier, operators are serialized
        as strings."""


class ValueModifier(Modifier):
    def __init__(self, value: Scalar, data_type: DataType) -> None:
        self.value = value
        self.data_type = data_type

    def to_dict(self, name: str) -> dict[str, Any]:
        return {
            "kind": "value",
            "name": name,
            "value": self.value,
            "dataType": self.data_type.value,
        }

    def to_value(self) -> Scalar:
        return self.value


class Enumeration(ValueModifier):
    def __init__(self, value: str) -> None:
        super().__init__(value, DataType.ENUMERATION)


def data_type_from_value(value: Scalar) -> DataType:
    if isinstance(value, bool):
        return DataType.BOOLEAN
    elif isinstance(value, int):
        return DataType.INTEGER
    elif isinstance(value, float):
        return DataType.REAL
    elif isinstance(value, str):
        return DataType.STRING

    raise ValueError(f"Unsupported type for modifier value {value}")


def ensure_as_modifier(value: Union[Scalar, Modifier]) -> Modifier:
    if isinstance(value, Modifier):
        return value

    return ValueModifier(value, data_type_from_value(value))


def modifiers_to_dict(variable_modifiers: dict[str, Modifier]) -> Any:
    return [m.to_dict(name) for name, m in variable_modifiers.items()]



"""Import service module."""
from typing import Any, Dict

from modelon.impact.client.sal.http import HTTPClient
from modelon.impact.client.sal.uri import URI


class ImportService:
    def __init__(self, uri: URI, http_client: HTTPClient):
        self._base_uri = uri
        self._http_client = http_client

    def get_import_status(self, location: str) -> Dict[str, Any]:
        url = (self._base_uri / location).resolve()
        return self._http_client.get_json(url)



from __future__ import annotations

from typing import TYPE_CHECKING, Any, Dict

from modelon.impact.client import exceptions
from modelon.impact.client.operations.base import (
    AsyncOperation,
    AsyncOperationStatus,
    Entity,
)

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import EntityFromOperation
    from modelon.impact.client.sal.service import Service


class CustomArtifactImportOperation(AsyncOperation[Entity]):
    """An operation class for the CustomArtifact class."""

    def __init__(
        self,
        location: str,
        workspace_id: str,
        exp_id: str,
        case_id: str,
        service: Service,
        create_entity: EntityFromOperation,
    ):
        super().__init__(create_entity)
        self._location = location
        self._workspace_id = workspace_id
        self._exp_id = exp_id
        self._case_id = case_id
        self._sal = service
        self._create_entity = create_entity

    def __repr__(self) -> str:
        return f"Content import operations for id '{self.id}'"

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, CustomArtifactImportOperation)
            and obj._location == self._location
        )

    @property
    def id(self) -> str:
        """Custom artifact import id."""
        return self._location.split("/")[-1]

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return "Custom artifact import"

    def _info(self) -> Dict[str, Any]:
        return self._sal.imports.get_import_status(self._location)["data"]

    def _get_artifact_download_name(self, artifact_id: str) -> str:
        resp = self._sal.experiment.case_artifacts_meta_get(
            self._workspace_id, self._exp_id, self._case_id
        )
        meta = next(
            (meta for meta in resp["data"]["items"] if meta["id"] == artifact_id),
            None,
        )
        if not meta:
            raise exceptions.NoSuchCustomArtifactError(
                f"No custom artifact found with ID: {artifact_id}."
            )
        return meta["downloadAs"]

    def data(self) -> Entity:
        """Returns a new CustomArtifact class instance.

        Returns:
            A CustomArtifact class instance.

        """
        info = self._info()
        if info["status"] == AsyncOperationStatus.ERROR.value:
            raise exceptions.IllegalCustomArtifactImport(
                f"CustomArtifact import failed! Cause: {info['error'].get('message')}"
            )
        artifact_id = info["data"]["artifactId"]
        download_as = self._get_artifact_download_name(artifact_id)
        return self._create_entity(
            self,
            workspace_id=self._workspace_id,
            experiment_id=self._exp_id,
            case_id=self._case_id,
            artifact_id=artifact_id,
            download_as=download_as,
        )

    @property
    def status(self) -> AsyncOperationStatus:
        """Returns the upload status as an enumeration.

        Returns:
            The AsyncOperationStatus enum. The status can have the enum values
            AsyncOperationStatus.READY, AsyncOperationStatus.RUNNING or
            AsyncOperationStatus.ERROR

        Example::

            case.import_custom_artifact('path/to/artifact.csv').status

        """
        return AsyncOperationStatus(self._info()["status"])



from __future__ import annotations

from abc import ABC, abstractmethod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from modelon.impact.client.sal.experiment import ExperimentService


class CustomArtifactInterface(ABC):
    def __init__(
        self,
        workspace_id: str,
        experiment_id: str,
        case_id: str,
        artifact_id: str,
        download_as: str,
        exp_sal: ExperimentService,
    ):
        self._workspace_id = workspace_id
        self._exp_id = experiment_id
        self._case_id = case_id
        self._artifact_id = artifact_id
        self._download_as = download_as
        self._exp_sal = exp_sal

    @property
    @abstractmethod
    def id(self) -> str:
        """Custom artifact id."""
        raise NotImplementedError



from __future__ import annotations

from typing import TYPE_CHECKING, Any

from modelon.impact.client.entities.interfaces.external_result import (
    ExternalResultInterface,
)
from modelon.impact.client.operations.external_result_import import (
    ExternalResultImportOperation,
)
from modelon.impact.client.sal.service import Service

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import BaseOperation


class ExternalResultMetaData:
    """Class containing external result metadata."""

    def __init__(self, id: str, name: str, description: str, workspace_id: str):
        self._id = id
        self._name = name
        self._description = description
        self._workspace_id = workspace_id

    @property
    def id(self) -> str:
        """Result id."""
        return self._id

    @property
    def name(self) -> str:
        """Label for result."""
        return self._name

    @property
    def description(self) -> str:
        """Description of the result."""
        return self._description

    @property
    def workspace_id(self) -> str:
        """Name of workspace."""
        return self._workspace_id


class ExternalResult(ExternalResultInterface):
    """Class containing  external result."""

    def __init__(self, result_id: str, service: Service):
        self._result_id = result_id
        self._sal = service

    def __repr__(self) -> str:
        return f"Result id '{self._result_id}'"

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, ExternalResult) and obj._result_id == self._result_id

    @property
    def id(self) -> str:
        """Result id."""
        return self._result_id

    @property
    def metadata(self) -> ExternalResultMetaData:
        """External result metadata."""
        upload_meta = self._sal.external_result.get_uploaded_result(self._result_id)[
            "data"
        ]
        id = upload_meta.get("id")
        name = upload_meta.get("name")
        description = upload_meta.get("description")
        workspace_id = upload_meta.get("workspaceId")
        return ExternalResultMetaData(id, name, description, workspace_id)

    def delete(self) -> None:
        self._sal.external_result.delete_uploaded_result(self._result_id)

    @classmethod
    def from_operation(
        cls, operation: BaseOperation[ExternalResult], **kwargs: Any
    ) -> ExternalResult:
        assert isinstance(operation, ExternalResultImportOperation)
        return cls(**kwargs, service=operation._sal)



from __future__ import annotations

import logging
import time
from typing import Any, Dict, List, Optional

from modelon.impact.client import exceptions
from modelon.impact.client.operations.base import AsyncOperationStatus
from modelon.impact.client.sal.service import Service

logger = logging.getLogger(__name__)


class OrphanPublishedWorkspaceOperation:
    """An operation class for the published workspace orphan cleanup."""

    def __init__(self, location: str, service: Service, operation_name: str):
        self._location = location
        self._sal = service
        self._name = operation_name

    def __repr__(self) -> str:
        return f"Orphan operations with id '{self.id}'"

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, OrphanPublishedWorkspaceOperation)
            and obj._location == self._location
        )

    @property
    def id(self) -> str:
        """Orphan operation id."""
        return self._location.split("/")[-1]

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return self._name

    def _info(self) -> Dict[str, Any]:
        return self._sal.imports.get_import_status(self._location)["data"]

    def data(self) -> List[str]:
        """Returns a list of orphaned published workspace IDs.

        Returns:
            A a list of orphaned published workspace IDs.

        """
        info = self._info()
        if info["status"] == AsyncOperationStatus.ERROR.value:
            raise exceptions.FailedOrphanCleanup(
                f"Orphan cleanup failed! Cause: {info['error'].get('message')}"
            )
        return info["data"]["orphans"]

    @property
    def status(self) -> AsyncOperationStatus:
        """Returns the operation status as an enumeration.

        Returns:
            The AsyncOperationStatus enum. The status can have the enum values
            AsyncOperationStatus.READY, AsyncOperationStatus.RUNNING or
            AsyncOperationStatus.ERROR

        Example::

            pw_client = client.get_published_workspaces_client()
            status = pw_client.wipe_orphans().status

        """
        return AsyncOperationStatus(self._info()["status"])

    def wait(self, timeout: Optional[float] = None) -> List[str]:
        """Waits until the operation completes. Returns the operation class instance if
        operation completes.

        Args:
            timeout: Time to wait in seconds for achieving the status. By default
                the timeout is set to 'None', which signifies an infinite time
                to wait until the status is achieved.

        Raises:
            OperationTimeOutError if time exceeds set timeout.

        Example::

            pw_client = client.get_published_workspaces_client()
            status = pw_client.wipe_orphans().wait()

        """
        start_t = time.time()
        while True:
            logger.info(f"{self.name} in progress! Status : {self.status.name}")
            if self.status.done():
                logger.info(f"{self.name} completed! Status : {self.status.name}")
                return self.data()

            current_t = time.time()
            if timeout and current_t - start_t > timeout:
                raise exceptions.OperationTimeOutError(
                    f"Time exceeded the set timeout - {timeout}s! "
                    f"Present status of operation is {self.status.name}!"
                )

            time.sleep(0.5)



from __future__ import annotations

import logging
from typing import TYPE_CHECKING, Any, Dict, Optional, Union

from modelon.impact.client import exceptions
from modelon.impact.client.operations.base import Entity, ExecutionOperation, Status

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import EntityFromOperation
    from modelon.impact.client.sal.service import Service

logger = logging.getLogger(__name__)


class CachedModelExecutableOperation(ExecutionOperation[Entity]):
    """An operation class for a cached ModelExecutable class."""

    def __init__(
        self,
        workspace_id: str,
        fmu_id: str,
        service: Service,
        create_entity: EntityFromOperation,
        info: Optional[Dict[str, Any]] = None,
        modifiers: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(create_entity)
        self._workspace_id = workspace_id
        self._fmu_id = fmu_id
        self._sal = service
        self._info = info
        self._modifiers = modifiers
        self._create_entity = create_entity

    def __repr__(self) -> str:
        return f"Cached model executable operations for id '{self._fmu_id}'"

    def __eq__(self, obj: object) -> bool:
        return (
            isinstance(obj, CachedModelExecutableOperation)
            and obj._fmu_id == self._fmu_id
        )

    @property
    def id(self) -> str:
        """FMU id."""
        return self._fmu_id

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return "Looking for cached FMU"

    def data(self) -> Entity:
        """Returns a new ModelExecutable class instance.

        Returns:
            A model_executable class instance.

        """
        return self._create_entity(
            self,
            workspace_id=self._workspace_id,
            fmu_id=self._fmu_id,
            info=self._info,
            modifiers=self._modifiers,
        )

    @property
    def status(self) -> Status:
        """Returns the compilation status as an enumeration.

        Returns:
            The compilation status enum. The status has the enum value
            Status.DONE as a cached FMU is only possible for a
            successfully finished compilation.

        Example::

            model.compile(options).status

        """
        return Status.DONE

    def cancel(self) -> None:
        raise NotImplementedError(
            "Cancel is not supported for CachedModelExecutableOperation class"
        )

    def wait(
        self,
        timeout: Optional[float] = None,
        status: Union[Status, tuple[Status, ...]] = (Status.DONE, Status.CANCELLED),
    ) -> Entity:
        """Waits until the operation achieves the set status. Returns the operation
        class instance if the set status is achieved.

        Args:
            timeout: Time to wait in seconds for achieving the status.This argument is
                not accounted for the CachedModelExecutableOperation class as the
                model has a successfully compiled status(Status.DONE) in this class.

            status:
                Operation status to be achieved. The only possible status for the
                CachedModelExecutableOperation class is Status.DONE as cached FMU
                is only available for a successfully compiled model.

        Returns:

            Entity class instance if the set status is achieved.

        Raises:

            OperationTimeOutError if operation status is not set as Status.DONE.

        Example::

           model.compile(compile_options).wait()

        """
        status_tuple = status if isinstance(status, tuple) else tuple([status])
        if self.status not in status_tuple:
            expected_status_names = [status.name for status in status_tuple]
            raise exceptions.OperationTimeOutError(
                current_status_name=self.status.name,
                expected_status_names=expected_status_names,
            )

        logger.info("Cached FMU found! Using the cached FMU!")
        return self.data()


class ModelExecutableOperation(ExecutionOperation[Entity]):
    """An operation class for the ModelExecutable class."""

    def __init__(
        self,
        workspace_id: str,
        fmu_id: str,
        service: Service,
        create_entity: EntityFromOperation,
    ):
        super().__init__(create_entity)
        self._workspace_id = workspace_id
        self._fmu_id = fmu_id
        self._sal = service
        self._create_entity = create_entity

    def __repr__(self) -> str:
        return f"Model executable operations for id '{self._fmu_id}'"

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, ModelExecutableOperation) and obj._fmu_id == self._fmu_id

    @property
    def id(self) -> str:
        """FMU id."""
        return self._fmu_id

    @property
    def name(self) -> str:
        """Return the name of operation."""
        return "Compilation"

    def data(self) -> Entity:
        """Returns a new ModelExecutable class instance.

        Returns:
            A ModelExecutable class instance.

        """
        return self._create_entity(
            self, workspace_id=self._workspace_id, fmu_id=self._fmu_id
        )

    @property
    def status(self) -> Status:
        """Returns the compilation status as an enumeration.

        Returns:
            The compilation status enum. The status can have the enum values
            Status.PENDING, Status.RUNNING, Status.STOPPING, Status.CANCELLED
            or Status.DONE

        Example::

            model.compile(options).status

        """
        return Status(
            self._sal.model_executable.compile_status(self._workspace_id, self._fmu_id)[
                "status"
            ]
        )

    def cancel(self) -> None:
        """Terminates the compilation process.

        Example::

            model.compile(options).cancel()

        """
        self._sal.model_executable.compile_cancel(self._workspace_id, self._fmu_id)



from __future__ import annotations

import enum
import logging
import warnings
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Union

from modelon.impact.client.entities.asserts import assert_variable_in_result
from modelon.impact.client.entities.case import Case
from modelon.impact.client.entities.custom_function import CustomFunction
from modelon.impact.client.entities.external_result import ExternalResult
from modelon.impact.client.entities.interfaces.experiment import ExperimentReference
from modelon.impact.client.entities.model import (
    Model,
    SimpleModelicaExperimentDefinition,
    to_domain_parameter_value,
)
from modelon.impact.client.entities.model_executable import (
    ModelExecutable,
    SimpleFMUExperimentDefinition,
)
from modelon.impact.client.entities.status import ExperimentStatus
from modelon.impact.client.experiment_definition.expansion import (
    ExpansionAlgorithm,
    FullFactorial,
    LatinHypercube,
    Sobol,
)
from modelon.impact.client.experiment_definition.extension import (
    SimpleExperimentExtension,
)
from modelon.impact.client.experiment_definition.operators import get_operator_from_dict
from modelon.impact.client.operations import experiment
from modelon.impact.client.options import (
    CompilerOptions,
    RuntimeOptions,
    SimulationOptions,
    SolverOptions,
)

if TYPE_CHECKING:
    from modelon.impact.client.operations.base import BaseOperation
    from modelon.impact.client.sal.service import Service

logger = logging.getLogger(__name__)

ScalarValue = Union[float, int, str]
ValidExperimentDefinitions = Union[
    SimpleModelicaExperimentDefinition,
    SimpleFMUExperimentDefinition,
]


@enum.unique
class _TrajectoryResponseFormat(enum.Enum):
    V1 = "application/vnd.impact.trajectories.v1+json"
    V2 = "application/vnd.impact.trajectories.v2+json"


class ExperimentResultPoint:
    """Value class with the single time instance data in a experiment."""

    def __init__(self, trajectories: List[Dict[str, Any]], variables: List[str]):
        self._trajectories = trajectories
        self._variables = variables

    @property
    def variables(self) -> List[str]:
        """List of variables."""
        return self._variables

    @property
    def cases(self) -> List[str]:
        """List of case ids."""
        return [case["caseId"] for case in self._trajectories]

    def as_lists(self) -> List[Optional[List[Optional[ScalarValue]]]]:
        """Return a list of results per case.

        None indicates that the case was not run or failed. None value
        for a variable indicated that the variable is not present in the
        specific case output.

        Example::

            result_data = result.as_lists()

        """
        return [
            [item["trajectory"][0] if item else None for item in case_data["items"]]
            for case_data in self._trajectories
        ]


@enum.unique
class _Workflow(enum.Enum):
    """Workflow type."""

    FMU_BASED = "FMU_BASED"
    CLASS_BASED = "CLASS_BASED"


class ExperimentRunInfo:
    """Class containing experiment run information."""

    def __init__(
        self,
        status: ExperimentStatus,
        errors: List[str],
        failed: int,
        successful: int,
        cancelled: int,
        not_started: int,
    ):
        self._status = status
        self._errors = errors
        self._failed = failed
        self._successful = successful
        self._cancelled = cancelled
        self._not_started = not_started

    @property
    def status(self) -> ExperimentStatus:
        """Status info for an Experiment.

        .. Note:: For tracing status changes of a running experiment execution,
            the ExperimentOperation status property must be used and not
            Experiment entities run info(ExperimentRunInfo) status property. See
            example for usage.

        Example::

            # To get the case entity run info status
            status = experiment.run_info.status

            # To get the status updates of a running experiment execution
            experiment_ops = experiment.execute()
            status = experiment_ops.status

        """
        return self._status

    @property
    def errors(self) -> List[str]:
        """A list of errors.

        Is empty unless 'status' attribute is 'FAILED'

        """
        return self._errors

    @property
    def successful(self) -> int:
        """Number of cases in experiment that are successful."""
        return self._successful

    @property
    def failed(self) -> int:
        """Number of cases in experiment thar have failed."""
        return self._failed

    @property
    def cancelled(self) -> int:
        """Number of cases in experiment that are cancelled."""
        return self._cancelled

    @property
    def not_started(self) -> int:
        """Number of cases in experiment that have not yet started."""
        return self._not_started


class ExperimentMetaData:
    """Class containing experiment metadata."""

    def __init__(self, meta_data: Dict[str, Any]):
        self._meta_data = meta_data

    @property
    def user_data(self) -> Dict[str, Any]:
        """User data dictionary object attached to experiment, if any."""
        return self._meta_data.get("user_data", {})

    @property
    def label(self) -> Optional[str]:
        """Experiment label."""
        return self._meta_data.get("label")


class Experiment(ExperimentReference):
    """Class containing Experiment functionalities."""

    def __init__(
        self,
        workspace_id: str,
        exp_id: str,
        service: Service,
        info: Optional[Dict[str, Any]] = None,
    ):
        super().__init__(workspace_id, exp_id, service, info)
        self._fmu_info: Optional[Dict[str, Any]] = None

    def __eq__(self, obj: object) -> bool:
        return isinstance(obj, Experiment) and obj._exp_id == self._exp_id

    def __repr__(self) -> str:
        return f"Experiment with id '{self._exp_id}'"

    def _get_info(self, cached: bool = True) -> Dict[str, Any]:
        if not cached or self._info is None:
            self._info = self._sal.workspace.experiment_get(
                self._workspace_id, self._exp_id
            )

        return self._info

    @property
    def run_info(self) -> ExperimentRunInfo:
        """Experiment run information.

        Returns:
            A ExperimentRunInfo class object.

        Example::

            status = experiment.run_info.status

        """
        run_info = self._get_info(cached=False)["run_info"]

        status = ExperimentStatus(run_info["status"])
        errors = run_info.get("errors", [])
        failed = run_info.get("failed", 0)
        successful = run_info.get("successful", 0)
        cancelled = run_info.get("cancelled", 0)
        not_started = run_info.get("not_started", 0)
        return ExperimentRunInfo(
            status, errors, failed, successful, cancelled, not_started
        )

    @property
    def metadata(self) -> ExperimentMetaData:
        """Experiment metadata.

        Returns:
            A ExperimentMetaData class object.

        Example::

            user_data = experiment.metadata.user_data

        """

        info = self._get_info(cached=False)
        meta_data = info.get("meta_data", {})
        return ExperimentMetaData(meta_data)

    @property
    def info(self) -> Dict[str, Any]:
        """Deprecated, use 'run_info' attribute."""
        logger.warning("This attribute is deprectated, use 'run_info' instead")
        return self._get_info(cached=False)

    def execute(
        self, with_cases: Optional[List[Case]] = None, sync_case_changes: bool = True
    ) -> experiment.ExperimentOperation:
        """Executes an experiment. Returns an ExperimentOperation class object.

        Args:
            with_cases: A list of cases objects to execute.
            sync_case_changes: Boolean specifying if to sync the cases given with the
                'with_cases' argument against the server before executing the
                experiment. Default is True.

        Returns:
            An ExperimentOperation class object.

        Example::

            experiment = workspace.create_experiment(experiment_definition)
            experiment_ops = experiment.execute()
            experiment_ops.cancel()
            experiment_ops.status
            experiment_ops.wait()

            generate_cases = experiment.execute(with_cases=[]).wait()
            cases_to_execute =  generate_cases.get_case('case_2')
            experiment = experiment.execute(with_cases=[cases_to_execute]).wait()

        """
        if sync_case_changes and with_cases is not None:
            for case in with_cases:
                case.sync()

        case_ids = [case.id for case in with_cases] if with_cases is not None else None
        return experiment.ExperimentOperation[Experiment](
            self._workspace_id,
            self._sal.experiment.experiment_execute(
                self._workspace_id, self._exp_id, case_ids
            ),
            self._sal,
            Experiment.from_operation,
        )

    def is_successful(self) -> bool:
        """Returns True if the Experiment is done and no cases has failed. Use the
        'run_info' attribute to get more info.

        Returns:
            True, if execution process has completed successfully. False, if
            execution process has failed, is cancelled or still running.

        Example::

            experiment.is_successful()

        """
        return (
            self.run_info.status == ExperimentStatus.DONE
            and self.run_info.failed == 0
            and self.run_info.cancelled == 0
        )

    def get_variables(self) -> List[str]:
        """Returns a list of variables available in the result.

        Returns:
            An list of result variables.

        Raises:
            OperationNotCompleteError if simulation process is in progress.
            OperationFailureError if simulation process has failed or was cancelled.

        Example::

            experiment.get_variables()

        """
        return self._sal.experiment.experiment_result_variables_get(
            self._workspace_id, self._exp_id
        )

    def get_cases(self) -> List[Case]:
        """Returns a list of case objects for an experiment.

        Returns:
            An list of case objects.

        Example::

            experiment.get_cases()

        """
        resp = self._sal.experiment.cases_get(self._workspace_id, self._exp_id)
        return [
            Case(case["id"], self._workspace_id, self._exp_id, self._sal, case)
            for case in resp["data"]["items"]
        ]

    def get_case(self, case_id: str) -> Case:
        """Returns a case object for a given case_id.

        Args:
            case_id: The case_id for the case.

        Returns:
            An case object.

        Example::

            experiment.get_case('case_1')

        """
        case_data = self._sal.experiment.case_get(
            self._workspace_id, self._exp_id, case_id
        )
        return Case(
            case_data["id"], self._workspace_id, self._exp_id, self._sal, case_data
        )

    def get_cases_with_label(self, case_label: str) -> List[Case]:
        """Returns a list of case objects for an experiment with the label.

        Args:
            case_label: The case_label for the case.

        Returns:
            An list of case objects.

        Example::

            experiment.get_cases_with_label('Cruise condition')

        """
        return [case for case in self.get_cases() if case.meta.label == case_label]

    def _validate_and_fetch_trajectories(
        self,
        variables: List[str],
        only_last_point: bool = False,
        format: _TrajectoryResponseFormat = _TrajectoryResponseFormat.V1,
    ) -> Any:
        if not isinstance(variables, list):
            raise TypeError(
                "Please specify the list of result keys for the trajectories of "
                "interest!"
            )
        assert_variable_in_result(variables, self.get_variables())

        return self._sal.experiment.trajectories_get(
            self._workspace_id, self._exp_id, variables, only_last_point, format.value
        )

    def get_trajectories(self, variables: List[str]) -> Dict[str, Any]:
        """Returns a dictionary containing the result trajectories for a list of result
        variables for all the cases.

        Args:
            variables: A list of result variables to fetch trajectories for.

        Returns:
            A dictionary object containing the result trajectories for all cases.

        Raises:
            TypeError if the variable is not a list object.
            ValueError if trajectory variable is not present in the result.

        Example::

            result = experiment.get_trajectories(['h', 'time'])
            height = result['case_1']['h']
            time = result['case_1']['time']

        """
        response = self._validate_and_fetch_trajectories(
            variables, only_last_point=False, format=_TrajectoryResponseFormat.V1
        )

        if response:
            case_nbrs = range(len(response[0]))
        else:
            case_nbrs = range(len(self.get_cases()))

        return {
            f"case_{j + 1}": {
                variable: response[i][j] for i, variable in enumerate(variables)
            }
            for j in case_nbrs
        }

    def get_last_point(
        self, variables: Optional[List[str]] = None
    ) -> ExperimentResultPoint:
        """Returns a ExperimentResultPoint class for a list of result variables for all
        the cases.

        Args:
            variables: An optional list of variables to fetch results for.

        Returns:
            An ExperimentResultPoint class object.

        Raises:
            TypeError if the variable is not a list object.
            ValueError if trajectory variable is not present in the result.

        Example::

            result = experiment.get_last_point(['h', 'time'])

            # Convert to Pandas data frame
            df = pd.DataFrame(data=result.as_lists(), columns=result.variables,
                index=result.cases)
            df.index.name = "Cases"

        """
        format = _TrajectoryResponseFormat.V2
        if variables is None:
            variables = self.get_variables()
            trajectories = self._sal.experiment.trajectories_get(
                self._workspace_id,
                self._exp_id,
                variables,
                last_point_only=True,
                format=format.value,
            )
        else:
            trajectories = self._validate_and_fetch_trajectories(
                variables, only_last_point=True, format=format
            )
        return ExperimentResultPoint(trajectories["data"]["items"], variables)

    def delete(self) -> None:
        """Deletes an experiment.

        Example::

            experiment.delete()

        """
        self._sal.experiment.experiment_delete(self._workspace_id, self._exp_id)

    @property
    def label(self) -> Optional[str]:
        """Returns the experiment label if it exists."""
        return self.metadata.label

    def set_label(self, label: str) -> None:
        """Sets a label (string) for an experiment to distinguish it. To get the label,
        you could call the experiment.label property.

        Example::

            experiment.set_label("Engine run with Oil type B")
            label = experiment.label

        """
        self._sal.experiment.experiment_set_label(
            self._workspace_id, self._exp_id, label
        )

    @classmethod
    def from_operation(
        cls, operation: BaseOperation[Experiment], **kwargs: Any
    ) -> Experiment:
        assert isinstance(operation, experiment.ExperimentOperation)
        return cls(**kwargs, service=operation._sal)

    @property
    def custom_function(self) -> str:
        """Returns the custom function name."""
        return self._get_info()["experiment"]["base"]["analysis"]["type"]

    def _get_fmu_info(self, fmu_id: str) -> Dict[str, Any]:
        if self._fmu_info is None:
            self._fmu_info = self._sal.workspace.fmu_get(self._workspace_id, fmu_id)

        return self._fmu_info

    def _get_workflow(self) -> _Workflow:
        model = self._get_info()["experiment"]["base"]["model"]
        return _Workflow.CLASS_BASED if model.get("modelica") else _Workflow.FMU_BASED

    def get_class_name(self) -> str:
        """Deprecated, use 'get_model_name' method."""
        warnings.warn(
            "The 'get_class_name' method is deprecated, use 'get_model_name' method"
            " instead!",
            category=DeprecationWarning,
            stacklevel=2,
        )
        return self.get_model_name()

    def get_model_name(self) -> str:
        """Return the model name."""
        model = self._get_info()["experiment"]["base"]["model"]
        if self._get_workflow() == _Workflow.CLASS_BASED:
            return model["modelica"]["className"]
        return self._get_fmu_info(model["fmu"]["id"])["input"]["class_name"]

    def get_compiler_options(self) -> CompilerOptions:
        """Return a CompilerOptions object."""
        model = self._get_info()["experiment"]["base"]["model"]
        if self._get_workflow() == _Workflow.CLASS_BASED:
            return CompilerOptions(
                model["modelica"].get("compilerOptions", {}), self.custom_function
            )
        fmu_info = self._get_fmu_info(model["fmu"]["id"])["input"]
        return CompilerOptions(
            fmu_info.get("compiler_options", {}), self.custom_function
        )

    def get_runtime_options(self) -> RuntimeOptions:
        """Return a RuntimeOptions object."""
        model = self._get_info()["experiment"]["base"]["model"]
        if self._get_workflow() == _Workflow.CLASS_BASED:
            return RuntimeOptions(
                model["modelica"].get("runtimeOptions", {}), self.custom_function
            )
        fmu_info = self._get_fmu_info(model["fmu"]["id"])["input"]
        return RuntimeOptions(fmu_info.get("runtime_options", {}), self.custom_function)

    def get_simulation_options(self) -> SimulationOptions:
        """Return a SimulationOptions object."""
        analysis = self._get_info()["experiment"]["base"]["analysis"]
        return SimulationOptions(
            analysis.get("simulationOptions", {}), self.custom_function
        )

    def get_solver_options(self) -> SolverOptions:
        """Return a SolverOptions object."""
        analysis = self._get_info()["experiment"]["base"]["analysis"]
        return SolverOptions(analysis.get("solverOptions", {}), self.custom_function)

    def _get_initialize_from_case(self, experiment_id: str, case_id: str) -> Case:
        case_data = self._sal.experiment.case_get(
            self._workspace_id, experiment_id, case_id
        )
        return Case(
            case_data["id"], self._workspace_id, experiment_id, self._sal, case_data
        )

    def _get_initialize_from_experiment(self, experiment_id: str) -> Experiment:
        resp = self._sal.workspace.experiment_get(self._workspace_id, experiment_id)
        return Experiment(self._workspace_id, resp["id"], self._sal, resp)

    def _get_initialize_from(
        self, modifiers: Dict[str, Any]
    ) -> Optional[Union[Case, Experiment, ExternalResult]]:
        if "initializeFrom" in modifiers:
            exp_id = modifiers["initializeFrom"]
            return self._get_initialize_from_experiment(exp_id)
        elif "initializeFromCase" in modifiers:
            case_id = modifiers["initializeFromCase"]["caseId"]
            exp_id = modifiers["initializeFromCase"]["experimentId"]
            return self._get_initialize_from_case(exp_id, case_id)
        elif "initializeFromExternalResult" in modifiers:
            ext_result_id = modifiers["initializeFromExternalResult"]
            return ExternalResult(result_id=ext_result_id, service=self._sal)
        return None

    def _get_extension_initialize_from(
        self, modifiers: Dict[str, Any]
    ) -> Optional[Union[Case, Experiment]]:
        if "initializeFrom" in modifiers:
            exp_id = modifiers["initializeFrom"]
            return self._get_initialize_from_experiment(exp_id)
        elif "initializeFromCase" in modifiers:
            case_id = modifiers["initializeFromCase"]["caseId"]
            exp_id = modifiers["initializeFromCase"]["experimentId"]
            return self._get_initialize_from_case(exp_id, case_id)
        return None

    def get_definition(self) -> ValidExperimentDefinitions:
        """Get an experiment definition that can be used to reproduce this experiment
        result.

        Returns:
            An instance of SimpleFMUExperimentDefinition or
            SimpleModelicaExperimentDefinition class.

        Example::

            definition = experiment.get_definition()

        """
        base = self._get_info(cached=False)["experiment"]["base"]
        analysis = base["analysis"]
        custom_function = self._get_custom_function(analysis)
        if self._get_workflow() == _Workflow.CLASS_BASED:
            model = Model(
                self.get_model_name(),
                workspace_id=self._workspace_id,
                project_id="",
                service=self._sal,
            )
            modelica = base["model"]["modelica"]
            definition = SimpleModelicaExperimentDefinition(
                model=model,
                custom_function=custom_function,
                compiler_options=self.get_compiler_options(),
                fmi_target=modelica["fmiTarget"],
                fmi_version=modelica["fmiVersion"],
                platform=modelica["platform"],
                compiler_log_level=modelica["compilerLogLevel"],
                runtime_options=self.get_runtime_options(),
                solver_options=self.get_solver_options(),
                simulation_options=self.get_simulation_options(),
                simulation_log_level=analysis["simulationLogLevel"],
                initialize_from=self._get_initialize_from(base["modifiers"]),
            )
            expansion = base.get("expansion", {})
            expansion = self._get_expansion_algorithm(
                expansion.get("algorithm"), expansion.get("parameters", {})
            )
            definition = definition.with_expansion(expansion=expansion)
        else:
            fmu_id = base["model"]["fmu"]["id"]
            definition = SimpleFMUExperimentDefinition(
                fmu=ModelExecutable(self._workspace_id, fmu_id, self._sal),
                custom_function=custom_function,
                solver_options=self.get_solver_options(),
                simulation_options=self.get_simulation_options(),
                simulation_log_level=analysis["simulationLogLevel"],
                initialize_from=self._get_initialize_from(base["modifiers"]),
            )  # type: ignore
        modifiers = {
            mod["name"]: get_operator_from_dict(mod)
            for mod in base["modifiers"]["variables"]
        }
        definition = definition.with_modifiers(modifiers=modifiers)
        extensions = self._get_info()["experiment"].get("extensions")
        if extensions:
            sim_exts = []
            for extension in extensions:
                analysis = extension.get("analysis", {})
                ext_custom_function_params = {
                    param["name"]: param["value"]
                    for param in analysis.get("parameters", [])
                }
                sim_ext = SimpleExperimentExtension(
                    parameter_modifiers=ext_custom_function_params,
                    solver_options=analysis.get("solverOptions"),
                    simulation_options=analysis.get("simulationOptions"),
                    simulation_log_level=analysis.get("simulationLogLevel"),
                    initialize_from=self._get_extension_initialize_from(
                        extension["modifiers"]
                    )
                    if extension.get("modifiers")
                    else None,
                )
                ext_modifiers = {
                    mod["name"]: to_domain_parameter_value(mod)
                    for mod in extension.get("modifiers", {}).get("variables", [])
                }
                sim_ext = sim_ext.with_modifiers(modifiers=ext_modifiers)
                case_data = extension.get("caseData", [])
                case_labels = [data.get("label") for data in case_data]
                if case_labels:
                    case_label = case_labels[0]
                    sim_ext = sim_ext.with_case_label(case_label)
                sim_exts.append(sim_ext)
            definition = definition.with_extensions(sim_exts)
        return definition

    def _get_custom_function(self, analysis: Dict[str, Any]) -> CustomFunction:
        custom_function_meta = self._sal.custom_function.custom_function_get(
            self._workspace_id, self.custom_function
        )
        custom_function_params = {
            param["name"]: param["value"] for param in analysis.get("parameters", [])
        }
        return CustomFunction(
            self._workspace_id,
            custom_function_meta["name"],
            custom_function_meta["parameters"],
            self._sal,
        ).with_parameters(**custom_function_params)

    def _get_expansion_algorithm(
        self, algorithm: str, parameters: Dict[str, Any]
    ) -> ExpansionAlgorithm:
        if algorithm == "SOBOL":
            return Sobol(samples=parameters["samples"])
        elif algorithm == "LATINHYPERCUBE":
            return LatinHypercube(
                samples=parameters["samples"], seed=parameters.get("seed")
            )
        else:
            return FullFactorial()

    @classmethod
    def from_reference(cls, reference: ExperimentReference) -> Experiment:
        return cls(
            workspace_id=reference._workspace_id,
            exp_id=reference._exp_id,
            service=reference._sal,
            info=reference._info,
        )
